"""Fixtures for testing.

This module provides custom fixtures for pytest that can be used to
automate common testing tasks and provide consistent setup and teardown
for tests.
"""

from collections.abc import Callable
from pathlib import Path

import pytest

from pyrig.dev.artifacts.builder.base.base import Builder
from pyrig.dev.configs.base.base import ConfigFile


@pytest.fixture
def config_file_factory[T: ConfigFile](
    tmp_path: Path,
) -> Callable[[type[T]], type[T]]:
    """Factory fixture for creating config file classes with tmp_path.

    This factory wraps any ConfigFile subclass to use tmp_path for get_path().
    Define tmp_path once here, then all test config classes inherit it.
    """

    def _make_test_config(
        base_class: type[T],
    ) -> type[T]:
        """Create a test config class that uses tmp_path."""

        class TestConfigFile(base_class):  # type: ignore [misc, valid-type]
            """Test config file with tmp_path override."""

            @classmethod
            def get_path(cls) -> Path:
                """Get the path to the config file in tmp_path."""
                path = super().get_path()
                return Path(tmp_path / path)

        return TestConfigFile

    return _make_test_config


@pytest.fixture
def builder_factory[T: Builder](tmp_path: Path) -> Callable[[type[T]], type[T]]:
    """Factory fixture for creating builder classes with tmp_path."""

    def _make_test_builder(base_class: type[T]) -> type[T]:
        """Create a test builder class that uses tmp_path."""

        class TestBuilder(base_class):  # type: ignore [misc, valid-type]
            """Test builder with tmp_path override."""

            @classmethod
            def get_artifacts_dir(cls) -> Path:
                """Get the artifacts directory."""
                return Path(tmp_path / cls.ARTIFACTS_DIR_NAME)

        return TestBuilder

    return _make_test_builder



"""Has the test for main in src."""

from pathlib import Path

from pyrig import main
from pyrig.dev.configs.base.base import PythonPackageConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.src.modules.module import to_path
from pyrig.src.testing.convention import make_test_obj_importpath_from_obj


class MainTestConfigFile(PythonPackageConfigFile):
    """Config file for test_main.py."""

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        test_module_path = to_path(
            make_test_obj_importpath_from_obj(main), is_package=False
        ).parent
        # replace pyrig with project name

        package_name = PyprojectConfigFile.get_package_name()
        test_module_path = Path(
            test_module_path.as_posix().replace("pyrig", package_name, 1)
        )
        return Path(test_module_path)

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return "test_main"

    @classmethod
    def get_content_str(cls) -> str:
        """Get the config."""
        return '''"""test module."""

from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.src.os.os import run_subprocess


def test_main() -> None:
    """Test func for main."""
    project_name = PyprojectConfigFile.get_project_name()
    stdout = run_subprocess(["poetry", "run", project_name, "--help"]).stdout.decode(
        "utf-8"
    )
    assert project_name in stdout
'''

    @classmethod
    def is_correct(cls) -> bool:
        """Check if the config is correct.

        Allow modifications to the test func.
        """
        return super().is_correct() or "def test_main" in cls.get_file_content()



"""Subcommands for the CLI.

They will be automatically imported and added to the CLI
IMPORTANT: All funcs in this file will be added as subcommands.
So best to define the logic elsewhere and just call it here in a wrapper.
"""

from pyrig.dev.artifacts.build import build as build_cmd
from pyrig.src.git.github.repo.protect import (
    protect_repository as protect_repo_cmd,
)
from pyrig.src.project.create_root import create_root as create_root_cmd
from pyrig.src.project.init import init as init_cmd
from pyrig.src.testing.create_tests import create_tests as create_tests_cmd


def create_root() -> None:
    """Creates the root of the project.

    This inits all ConfigFiles and creates __init__.py files for the src
    and tests package where they are missing. It does not overwrite any
    existing files.
    """
    create_root_cmd()


def create_tests() -> None:
    """Create all test files for the project.

    This generates test skeletons for all functions and classes in the src
    package. It does not overwrite any existing tests.
    Tests are also automatically generated when missing by running pytest.
    """
    create_tests_cmd()


def init() -> None:
    """Set up the project.

    This is the setup command when you created the project from scratch.
    It will init all config files, create the root, create tests, and run
    all pre-commit hooks and tests.
    """
    init_cmd()


def build() -> None:
    """Build all artifacts.

    Invokes every subclass of Builder in the builder package.
    """
    build_cmd()


def protect_repo() -> None:
    """Protect the repository.

    This will set secure repo settings and add a branch protection rulesets.
    """
    protect_repo_cmd()



"""Config File subclass that creates the builds dir and a build.py."""

from types import ModuleType

from pyrig.dev.artifacts.builder import builder
from pyrig.dev.configs.base.base import CopyModuleConfigFile


class BuilderConfigFile(CopyModuleConfigFile):
    """Config File subclass that creates the dirs folder."""

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return builder



"""Config utilities for main.py."""

from types import ModuleType

from pyrig import main
from pyrig.dev.configs.base.base import CopyModuleConfigFile


class MainConfigFile(CopyModuleConfigFile):
    """Config file for main.py.

    Creates a main.py in pkg_name/src.
    """

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return main

    @classmethod
    def is_correct(cls) -> bool:
        """Check if the config is correct.

        Allow modifications to the main func and __name__ == '__main__' line.
        """
        return super().is_correct() or (
            "def main" in cls.get_file_content()
            and 'if __name__ == "__main__":' in cls.get_file_content()
        )



"""Config utilities for resources/__init__.py."""

from pathlib import Path
from types import ModuleType

from pyrig.dev.artifacts import resources
from pyrig.dev.configs.base.base import CopyModuleConfigFile
from pyrig.src.modules.module import get_isolated_obj_name


class ResourcesInitConfigFile(CopyModuleConfigFile):
    """Config file for resources/__init__.py."""

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return resources

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        path = super().get_parent_path()
        # this path will be parent of resources
        return path / get_isolated_obj_name(resources)

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return "__init__"



"""Main entrypoint for the project."""

import pyrig


def main() -> None:
    """Main entrypoint for the project."""
    msg = f"""Add your projects entrypoint code to this function.
This function is automatically added to your cli by {pyrig.__name__}.
You can call it with
`poetry run your-pkg-name main`
or via
`python -m your-pkg-name`.
"""
    raise NotImplementedError(msg)


if __name__ == "__main__":
    main()



"""Testing conventions and utilities.

This module provides functions and constants for managing test naming conventions,
mapping between test objects and their corresponding implementation objects,
and utilities for test discovery and validation.

Returns:
    Various utility functions and constants for testing conventions.

"""

from collections.abc import Callable, Iterable
from types import ModuleType
from typing import Any

from pyrig.src.modules.module import (
    get_isolated_obj_name,
    import_obj_from_importpath,
    make_obj_importpath,
)

TEST_FUNCTION_PREFIX = "test_"

TEST_CLASS_PREFIX = "Test"

TEST_MODULE_PREFIX = TEST_FUNCTION_PREFIX

TEST_PREFIXES = [TEST_FUNCTION_PREFIX, TEST_CLASS_PREFIX, TEST_MODULE_PREFIX]

TESTS_PACKAGE_NAME = "tests"


def get_right_test_prefix(obj: Callable[..., Any] | type | ModuleType) -> str:
    """Get the appropriate test prefix for an object based on its type.

    Args:
        obj: The object to get the test prefix for (function, class, or module)

    Returns:
        The appropriate test prefix string for the given object type

    """
    if isinstance(obj, ModuleType):
        return TEST_MODULE_PREFIX
    if isinstance(obj, type):
        return TEST_CLASS_PREFIX
    return TEST_FUNCTION_PREFIX


def make_test_obj_name(obj: Callable[..., Any] | type | ModuleType) -> str:
    """Create a test name for an object by adding the appropriate prefix.

    Args:
        obj: The object to create a test name for

    Returns:
        The test name with the appropriate prefix

    """
    prefix = get_right_test_prefix(obj)
    name = get_isolated_obj_name(obj)
    return prefix + name


def reverse_make_test_obj_name(test_name: str) -> str:
    """Extract the original object name from a test name by removing the prefix.

    Args:
        test_name: The test name to extract the original name from

    Returns:
        The original object name without the test prefix

    Raises:
        ValueError: If the test name doesn't start with any of the expected prefixes

    """
    for prefix in TEST_PREFIXES:
        if test_name.startswith(prefix):
            return test_name.removeprefix(prefix)
    msg = f"{test_name=} is expected to start with one of {TEST_PREFIXES=}"
    raise ValueError(msg)


def make_test_obj_importpath_from_obj(
    obj: Callable[..., Any] | type | ModuleType,
) -> str:
    """Create an import path for a test object based on the original object.

    Args:
        obj: The original object to create a test import path for

    Returns:
        The import path for the corresponding test object

    """
    parts = make_obj_importpath(obj).split(".")
    test_name = make_test_obj_name(obj)
    test_parts = [
        (TEST_MODULE_PREFIX if part[0].islower() else TEST_CLASS_PREFIX) + part
        for part in parts
    ]
    test_parts[-1] = test_name
    test_parts.insert(0, TESTS_PACKAGE_NAME)
    return ".".join(test_parts)


def make_obj_importpath_from_test_obj(
    test_obj: Callable[..., Any] | type | ModuleType,
) -> str:
    """Create an import path for an original object based on its test object.

    Args:
        test_obj: The test object to create an original import path for

    Returns:
        The import path for the corresponding original object

    """
    test_parts = make_obj_importpath(test_obj).split(".")
    test_parts = test_parts[1:]
    parts = [reverse_make_test_obj_name(part) for part in test_parts]
    return ".".join(parts)


def get_test_obj_from_obj(
    obj: Callable[..., Any] | type | ModuleType,
) -> Callable[..., Any] | type | ModuleType:
    """Get the test object corresponding to an original object.

    Args:
        obj: The original object to get the test object for

    Returns:
        The corresponding test object

    """
    test_obj_path = make_test_obj_importpath_from_obj(obj)
    return import_obj_from_importpath(test_obj_path)


def get_obj_from_test_obj(
    test_obj: Callable[..., Any] | type | ModuleType,
) -> Callable[..., Any] | type | ModuleType:
    """Get the original object corresponding to a test object.

    Args:
        test_obj: The test object to get the original object for

    Returns:
        The corresponding original object

    """
    obj_importpath = make_obj_importpath_from_test_obj(test_obj)
    return import_obj_from_importpath(obj_importpath)


def make_untested_summary_error_msg(
    untested_objs: Iterable[str],
) -> str:
    """Create an error message summarizing untested objects.

    Args:
        untested_objs: Collection of import paths for untested objects

    Returns:
        A formatted error message listing all untested objects

    """
    msg = """
    Found untested objects:
    """
    for untested in untested_objs:
        msg += f"""
        - {untested}
        """
    return msg



"""String utilities."""

import re
from collections.abc import Callable
from types import ModuleType
from typing import Any


def split_on_uppercase(string: str) -> list[str]:
    """Split a string on uppercase letters.

    Args:
        string: String to split

    Returns:
        List of substrings split on uppercase letters

    Example:
        split_on_uppercase("HelloWorld") -> ["Hello", "World"]

    """
    return [s for s in re.split(r"(?=[A-Z])", string) if s]


def make_name_from_obj(
    package: ModuleType | Callable[..., Any] | type | str,
    split_on: str = "_",
    join_on: str = "-",
    *,
    capitalize: bool = True,
) -> str:
    """Make a name from a package.

    takes a package and makes a name from it that is readable by humans.

    Args:
        package (ModuleType): The package to make a name from
        split_on (str, optional): what to split the package name on. Defaults to "_".
        join_on (str, optional): what to join the package name with. Defaults to "-".
        capitalize (bool, optional): Whether to capitalize each part. Defaults to True.

    Returns:
        str: _description_
    """
    if not isinstance(package, str):
        package_name = package.__name__.split(".")[-1]
    else:
        package_name = package
    parts = package_name.split(split_on)
    if capitalize:
        parts = [part.capitalize() for part in parts]
    return join_on.join(parts)



"""Iterating utilities for handling iterables.

This module provides utility functions for working with iterables,
including getting the length of an iterable with a default value.
These utilities help with iterable operations and manipulations.
"""

from collections.abc import Callable, Iterable
from typing import Any


def nested_structure_is_subset(
    subset: dict[Any, Any] | list[Any] | Any,
    superset: dict[Any, Any] | list[Any] | Any,
    on_false_dict_action: Callable[[dict[Any, Any], dict[Any, Any], Any], Any]
    | None = None,
    on_false_list_action: Callable[[list[Any], list[Any], int], Any] | None = None,
) -> bool:
    """Check if a dictionary is a nested subset of another dictionary.

    Args:
        subset: Dictionary to check
        superset: Dictionary to check against
        on_false_dict_action: Action to take on each false dict comparison
            must return a bool to indicate if after action is still false
        on_false_list_action: Action to take on each false list comparison
            must return a bool to indicate if after action is still false

    Each value of a key must be equal to the value of the same key in the superset.
    If the value is dictionary, the function is called recursively.
    If the value is list, each item must be in the list of the same key in the superset.
    The order in lists does not matter.

    Returns:
        True if subset is a nested subset of superset, False otherwise
    """
    if isinstance(subset, dict) and isinstance(superset, dict):
        iterable: Iterable[tuple[Any, Any]] = subset.items()
        on_false_action: Callable[[Any, Any, Any], Any] | None = on_false_dict_action

        def get_actual(key_or_index: Any) -> Any:
            """Get actual value from superset."""
            return superset.get(key_or_index)

    elif isinstance(subset, list) and isinstance(superset, list):
        iterable = enumerate(subset)
        on_false_action = on_false_list_action

        def get_actual(key_or_index: Any) -> Any:
            """Get actual value from superset."""
            subset_val = subset[key_or_index]
            for superset_val in superset:
                if nested_structure_is_subset(subset_val, superset_val):
                    return superset_val

            return superset[key_or_index] if key_or_index < len(superset) else None
    else:
        return subset == superset

    all_good = True
    for key_or_index, value in iterable:
        actual_value = get_actual(key_or_index)
        if not nested_structure_is_subset(
            value, actual_value, on_false_dict_action, on_false_list_action
        ):
            all_good = False
            if on_false_action is not None:
                on_false_action(subset, superset, key_or_index)
                all_good = nested_structure_is_subset(subset, superset)

    return all_good



"""Testing fixtures for pytest.

This module provides custom fixtures for pytest that can be used to
automate common testing tasks and provide consistent setup and teardown
for tests.
"""

import functools

import pytest

function_fixture = functools.partial(pytest.fixture, scope="function")
class_fixture = functools.partial(pytest.fixture, scope="class")
module_fixture = functools.partial(pytest.fixture, scope="module")
package_fixture = functools.partial(pytest.fixture, scope="package")
session_fixture = functools.partial(pytest.fixture, scope="session")

autouse_function_fixture = functools.partial(
    pytest.fixture, scope="function", autouse=True
)
autouse_class_fixture = functools.partial(pytest.fixture, scope="class", autouse=True)
autouse_module_fixture = functools.partial(pytest.fixture, scope="module", autouse=True)
autouse_package_fixture = functools.partial(
    pytest.fixture, scope="package", autouse=True
)
autouse_session_fixture = functools.partial(
    pytest.fixture, scope="session", autouse=True
)



"""Contains base utilities for git workflows."""

from abc import abstractmethod
from collections.abc import Callable
from pathlib import Path
from typing import Any, ClassVar

import pyrig
from pyrig.dev.artifacts.builder.base.base import Builder
from pyrig.dev.cli.subcommands import build, protect_repo
from pyrig.dev.configs.base.base import YamlConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.src.modules.package import DependencyGraph, get_src_package
from pyrig.src.project.poetry.poetry import get_poetry_run_pyrig_cli_cmd_script
from pyrig.src.string import (
    make_name_from_obj,
    split_on_uppercase,
)


class Workflow(YamlConfigFile):
    """Base class for workflows."""

    UBUNTU_LATEST = "ubuntu-latest"
    WINDOWS_LATEST = "windows-latest"
    MACOS_LATEST = "macos-latest"

    ARTIFACTS_DIR_NAME = Builder.ARTIFACTS_DIR_NAME
    ARTIFACTS_PATTERN = f"{ARTIFACTS_DIR_NAME}/*"

    EMPTY_CONFIG: ClassVar[dict[str, Any]] = {
        "on": {
            "workflow_dispatch": {},
        },
        "jobs": {
            "empty": {
                "runs-on": "ubuntu-latest",
                "steps": [
                    {
                        "name": "Empty Step",
                        "run": "echo 'Empty Step'",
                    }
                ],
            },
        },
    }

    @classmethod
    def get_configs(cls) -> dict[str, Any]:
        """Get the workflow config."""
        return {
            "name": cls.get_workflow_name(),
            "on": cls.get_workflow_triggers(),
            "permissions": cls.get_permissions(),
            "run-name": cls.get_run_name(),
            "defaults": cls.get_defaults(),
            "env": cls.get_global_env(),
            "jobs": cls.get_jobs(),
        }

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path(".github/workflows")

    @classmethod
    def is_correct(cls) -> bool:
        """Check if the config is correct.

        Needs some special handling since workflow files cannot be empty.
        We need a workflow that will never trigger and even if doesnt do anything.
        """
        correct = super().is_correct()
        if cls.get_path().read_text(encoding="utf-8") == "":
            # dump a dispatch in there for on and an empty job for jobs
            cls.dump(cls.EMPTY_CONFIG)

        return correct or cls.load() == cls.EMPTY_CONFIG

    # Overridable Workflow Parts
    # ----------------------------------------------------------------------------
    @classmethod
    @abstractmethod
    def get_jobs(cls) -> dict[str, Any]:
        """Get the workflow jobs."""

    @classmethod
    def get_workflow_triggers(cls) -> dict[str, Any]:
        """Get the workflow triggers.

        Can be overriden. Standard is workflow_dispatch.
        """
        return cls.on_workflow_dispatch()

    @classmethod
    def get_permissions(cls) -> dict[str, Any]:
        """Get the workflow permissions. Can be overriden.

        Standard is no extra permissions.
        """
        return {}

    @classmethod
    def get_defaults(cls) -> dict[str, Any]:
        """Get the workflow defaults. Can be overriden.

        Standard is bash.
        """
        return {"run": {"shell": "bash"}}

    @classmethod
    def get_global_env(cls) -> dict[str, Any]:
        """Get the global env. Can be overriden.

        Standard is no global env.
        """
        return {"PYTHONDONTWRITEBYTECODE": 1}

    # Workflow Conventions
    # ----------------------------------------------------------------------------
    @classmethod
    def get_workflow_name(cls) -> str:
        """Get the workflow name."""
        return " ".join(split_on_uppercase(cls.__name__))

    @classmethod
    def get_run_name(cls) -> str:
        """Get the run name."""
        return cls.get_workflow_name()

    # Build Utilities
    # ----------------------------------------------------------------------------
    @classmethod
    def get_job(  # noqa: PLR0913
        cls,
        job_func: Callable[..., Any],
        needs: list[str] | None = None,
        strategy: dict[str, Any] | None = None,
        permissions: dict[str, Any] | None = None,
        runs_on: str = UBUNTU_LATEST,
        if_condition: str | None = None,
        steps: list[dict[str, Any]] | None = None,
        job: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get a job.

        Args:
        job_func: The function that represents the job. Used to generate the name.
        job: The job to update. Defaults to a new job.
        needs: The needs of the job.
        strategy: The strategy of the job. like matrix
        permissions: The permissions of the job.
        runs_on: The runs-on of the job. Defaults to ubuntu-latest.
        if_condition: The if condition of the job.
        steps: The steps of the job.

        Returns:
        The job.
        """
        name = cls.make_id_from_func(job_func)
        if job is None:
            job = {}
        job_config: dict[str, Any] = {}
        if needs is not None:
            job_config["needs"] = needs
        if strategy is not None:
            job_config["strategy"] = strategy
        if permissions is not None:
            job_config["permissions"] = permissions
        job_config["runs-on"] = runs_on
        if if_condition is not None:
            job_config["if"] = if_condition
        if steps is not None:
            job_config["steps"] = steps
        job_config.update(job)
        return {name: job_config}

    @classmethod
    def make_name_from_func(cls, func: Callable[..., Any]) -> str:
        """Make a name from a function."""
        name = make_name_from_obj(func, split_on="_", join_on=" ", capitalize=True)
        prefix = split_on_uppercase(name)[0]
        return name.removeprefix(prefix).strip()

    @classmethod
    def make_id_from_func(cls, func: Callable[..., Any]) -> str:
        """Make an id from a function."""
        name = func.__name__
        prefix = name.split("_")[0]
        return name.removeprefix(f"{prefix}_")

    # triggers
    @classmethod
    def on_workflow_dispatch(cls) -> dict[str, Any]:
        """Get the workflow dispatch trigger."""
        return {"workflow_dispatch": {}}

    @classmethod
    def on_push(cls, branches: list[str] | None = None) -> dict[str, Any]:
        """Get the push trigger."""
        if branches is None:
            branches = ["main"]
        return {"push": {"branches": branches}}

    @classmethod
    def on_schedule(cls, cron: str) -> dict[str, Any]:
        """Get the schedule trigger."""
        return {"schedule": [{"cron": cron}]}

    @classmethod
    def on_pull_request(cls, types: list[str] | None = None) -> dict[str, Any]:
        """Get the pull request trigger."""
        if types is None:
            types = ["opened", "synchronize", "reopened"]
        return {"pull_request": {"types": types}}

    @classmethod
    def on_workflow_run(cls, workflows: list[str] | None = None) -> dict[str, Any]:
        """Get the workflow run trigger."""
        if workflows is None:
            workflows = [cls.get_workflow_name()]
        return {"workflow_run": {"workflows": workflows, "types": ["completed"]}}

    # permissions
    @classmethod
    def permission_content(cls, permission: str = "read") -> dict[str, Any]:
        """Get the content read permission."""
        return {"contents": permission}

    # Steps
    @classmethod
    def get_step(  # noqa: PLR0913
        cls,
        step_func: Callable[..., Any],
        run: str | None = None,
        if_condition: str | None = None,
        uses: str | None = None,
        with_: dict[str, Any] | None = None,
        env: dict[str, Any] | None = None,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get a step.

        Args:
        step_func: The function that represents the step. Used to generate the name.
        run: The run command.
        if_condition: The if condition.
        uses: The uses command.
        with_: The with command.
        env: The env command.
        step: The step to update. Defaults to a new step.

        Returns:
        The step.
        """
        if step is None:
            step = {}
        # make name from setup function name if name is a function
        name = cls.make_name_from_func(step_func)
        id_ = cls.make_id_from_func(step_func)
        step_config: dict[str, Any] = {"name": name, "id": id_}
        if run is not None:
            step_config["run"] = run
        if if_condition is not None:
            step_config["if"] = if_condition
        if uses is not None:
            step_config["uses"] = uses
        if with_ is not None:
            step_config["with"] = with_
        if env is not None:
            step_config["env"] = env

        step_config.update(step)

        return step_config

    # Strategy
    @classmethod
    def strategy_matrix_os_and_python_version(
        cls,
        os: list[str] | None = None,
        python_version: list[str] | None = None,
        matrix: dict[str, list[Any]] | None = None,
        strategy: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get a strategy for os and python version."""
        return cls.strategy_matrix(
            matrix=cls.matrix_os_and_python_version(
                os=os, python_version=python_version, matrix=matrix
            ),
            strategy=strategy,
        )

    @classmethod
    def strategy_matrix_python_version(
        cls,
        python_version: list[str] | None = None,
        matrix: dict[str, list[Any]] | None = None,
        strategy: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get a strategy for python version."""
        return cls.strategy_matrix(
            matrix=cls.matrix_python_version(
                python_version=python_version, matrix=matrix
            ),
            strategy=strategy,
        )

    @classmethod
    def strategy_matrix_os(
        cls,
        os: list[str] | None = None,
        matrix: dict[str, list[Any]] | None = None,
        strategy: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get a strategy for os."""
        return cls.strategy_matrix(
            matrix=cls.matrix_os(os=os, matrix=matrix), strategy=strategy
        )

    @classmethod
    def strategy_matrix(
        cls,
        *,
        strategy: dict[str, Any] | None = None,
        matrix: dict[str, list[Any]] | None = None,
    ) -> dict[str, Any]:
        """Get a matrix strategy."""
        if strategy is None:
            strategy = {}
        if matrix is None:
            matrix = {}
        strategy["matrix"] = matrix
        return cls.get_strategy(strategy=strategy)

    @classmethod
    def get_strategy(
        cls,
        *,
        strategy: dict[str, Any],
    ) -> dict[str, Any]:
        """Get a strategy."""
        strategy["fail-fast"] = strategy.pop("fail-fast", True)
        return strategy

    @classmethod
    def matrix_os_and_python_version(
        cls,
        os: list[str] | None = None,
        python_version: list[str] | None = None,
        matrix: dict[str, list[Any]] | None = None,
    ) -> dict[str, Any]:
        """Get a matrix for os and python version."""
        if matrix is None:
            matrix = {}
        os_matrix = cls.matrix_os(os=os, matrix=matrix)["os"]
        python_version_matrix = cls.matrix_python_version(
            python_version=python_version, matrix=matrix
        )["python-version"]
        matrix["os"] = os_matrix
        matrix["python-version"] = python_version_matrix
        return cls.get_matrix(matrix=matrix)

    @classmethod
    def matrix_os(
        cls,
        *,
        os: list[str] | None = None,
        matrix: dict[str, list[Any]] | None = None,
    ) -> dict[str, Any]:
        """Get a matrix for os."""
        if os is None:
            os = [cls.UBUNTU_LATEST, cls.WINDOWS_LATEST, cls.MACOS_LATEST]
        if matrix is None:
            matrix = {}
        matrix["os"] = os
        return cls.get_matrix(matrix=matrix)

    @classmethod
    def matrix_python_version(
        cls,
        *,
        python_version: list[str] | None = None,
        matrix: dict[str, list[Any]] | None = None,
    ) -> dict[str, Any]:
        """Get a matrix for python version."""
        if python_version is None:
            python_version = [
                str(v) for v in PyprojectConfigFile.get_supported_python_versions()
            ]
        if matrix is None:
            matrix = {}
        matrix["python-version"] = python_version
        return cls.get_matrix(matrix=matrix)

    @classmethod
    def get_matrix(cls, matrix: dict[str, list[Any]]) -> dict[str, Any]:
        """Get a matrix."""
        return matrix

    # Workflow Steps
    # ----------------------------------------------------------------------------
    # Combined Steps
    @classmethod
    def steps_core_setup(
        cls, python_version: str | None = None, *, repo_token: bool = False
    ) -> list[dict[str, Any]]:
        """Get the core setup steps."""
        return [
            cls.step_checkout_repository(repo_token=repo_token),
            cls.step_setup_python(python_version=python_version),
            cls.step_setup_poetry(),
        ]

    @classmethod
    def steps_core_installed_setup(
        cls, python_version: str | None = None, *, repo_token: bool = False
    ) -> list[dict[str, Any]]:
        """Get the core setup steps."""
        return [
            *cls.steps_core_setup(python_version=python_version, repo_token=repo_token),
            cls.step_add_poetry_to_windows_path(),
            cls.step_install_python_dependencies(),
            *cls.steps_configure_keyring_if_needed(),
            cls.step_update_dependencies(),
            cls.step_add_dependency_updates_to_git(),
        ]

    @classmethod
    def steps_core_matrix_setup(
        cls, python_version: str | None = None, *, repo_token: bool = False
    ) -> list[dict[str, Any]]:
        """Get the core matrix setup steps."""
        return [
            *cls.steps_core_installed_setup(
                python_version=python_version, repo_token=repo_token
            ),
        ]

    @classmethod
    def steps_configure_keyring_if_needed(cls) -> list[dict[str, Any]]:
        """Get the keyring steps if keyring is in dependencies."""
        steps: list[dict[str, Any]] = []
        if "keyring" in DependencyGraph.get_all_dependencies():
            steps.append(cls.step_setup_keyring())
        return steps

    # Single Step
    @classmethod
    def step_aggregate_matrix_results(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the aggregate matrix results step."""
        return cls.get_step(
            step_func=cls.step_aggregate_matrix_results,
            run="echo 'Aggregating matrix results into one job.'",
            step=step,
        )

    @classmethod
    def step_no_builder_defined(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the no build script step."""
        return cls.get_step(
            step_func=cls.step_no_builder_defined,
            run="echo 'No non-abstract builders defined. Skipping build.'",
            step=step,
        )

    @classmethod
    def step_run_tests(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the run tests step."""
        if step is None:
            step = {}
        if PyprojectConfigFile.get_package_name() == pyrig.__name__:
            step.setdefault("env", {})["REPO_TOKEN"] = cls.insert_repo_token()
        return cls.get_step(
            step_func=cls.step_run_tests,
            run="poetry run pytest --log-cli-level=INFO",
            step=step,
        )

    @classmethod
    def step_patch_version(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the patch version step."""
        return cls.get_step(
            step_func=cls.step_patch_version,
            run="poetry version patch && git add pyproject.toml",
            step=step,
        )

    @classmethod
    def step_update_dependencies(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the update dependencies step."""
        return cls.get_step(
            step_func=cls.step_update_dependencies,
            run="poetry update --with dev",
            step=step,
        )

    @classmethod
    def step_add_dependency_updates_to_git(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the add dependency updates to git step."""
        return cls.get_step(
            step_func=cls.step_add_dependency_updates_to_git,
            run="git add pyproject.toml poetry.lock",
            step=step,
        )

    @classmethod
    def step_checkout_repository(
        cls,
        *,
        step: dict[str, Any] | None = None,
        fetch_depth: int | None = None,
        repo_token: bool = False,
    ) -> dict[str, Any]:
        """Get the checkout step."""
        if step is None:
            step = {}
        if fetch_depth is not None:
            step.setdefault("with", {})["fetch-depth"] = fetch_depth
        if repo_token:
            step.setdefault("with", {})["token"] = cls.insert_repo_token()
        return cls.get_step(
            step_func=cls.step_checkout_repository,
            uses="actions/checkout@main",
            step=step,
        )

    @classmethod
    def step_setup_git(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the setup git step."""
        return cls.get_step(
            step_func=cls.step_setup_git,
            run='git config --global user.email "github-actions[bot]@users.noreply.github.com" && git config --global user.name "github-actions[bot]"',  # noqa: E501
            step=step,
        )

    @classmethod
    def step_setup_python(
        cls,
        *,
        step: dict[str, Any] | None = None,
        python_version: str | None = None,
    ) -> dict[str, Any]:
        """Get the setup python step."""
        if step is None:
            step = {}
        if python_version is None:
            python_version = str(
                PyprojectConfigFile.get_latest_possible_python_version()
            )

        step.setdefault("with", {})["python-version"] = python_version
        return cls.get_step(
            step_func=cls.step_setup_python,
            uses="actions/setup-python@main",
            step=step,
        )

    @classmethod
    def step_setup_poetry(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the setup poetry step."""
        return cls.get_step(
            step_func=cls.step_setup_poetry,
            uses="snok/install-poetry@main",
            step=step,
        )

    @classmethod
    def step_add_poetry_to_windows_path(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the add poetry to path step."""
        return cls.get_step(
            step_func=cls.step_add_poetry_to_windows_path,
            run="echo 'C:/Users/runneradmin/.local/bin' >> $GITHUB_PATH",
            if_condition=f"{cls.insert_os()} == 'Windows'",
            step=step,
        )

    @classmethod
    def step_add_pypi_token_to_poetry(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the add pypi token to poetry step."""
        return cls.get_step(
            step_func=cls.step_add_pypi_token_to_poetry,
            run="poetry config pypi-token.pypi ${{ secrets.PYPI_TOKEN }}",
            step=step,
        )

    @classmethod
    def step_publish_to_pypi(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the publish to pypi step."""
        return cls.get_step(
            step_func=cls.step_publish_to_pypi,
            run="poetry publish --build",
            step=step,
        )

    @classmethod
    def step_install_python_dependencies(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the install dependencies step."""
        return cls.get_step(
            step_func=cls.step_install_python_dependencies,
            run="poetry install --with dev",
            step=step,
        )

    @classmethod
    def step_setup_keyring(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the setup keyring step."""
        return cls.get_step(
            step_func=cls.step_setup_keyring,
            run='poetry run pip install keyrings.alt && poetry run python -c "import keyring; from keyrings.alt.file import PlaintextKeyring; keyring.set_keyring(PlaintextKeyring());"',  # noqa: E501
            step=step,
        )

    @classmethod
    def step_protect_repository(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the protect repository step."""
        return cls.get_step(
            step_func=cls.step_protect_repository,
            run=get_poetry_run_pyrig_cli_cmd_script(
                cmd=protect_repo,
            ),
            env={"REPO_TOKEN": cls.insert_repo_token()},
            step=step,
        )

    @classmethod
    def step_run_pre_commit_hooks(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the run pre-commit hooks step.

        Patching version is useful to have at least a minimal version bump when
        creating a release and it also makes sure git stash pop does not fail when
        there are no changes.
        """
        return cls.get_step(
            step_func=cls.step_run_pre_commit_hooks,
            run="poetry run pre-commit run --all-files",
            step=step,
        )

    @classmethod
    def step_add_version_patch(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the add version patch step."""
        return cls.get_step(
            step_func=cls.step_add_version_patch,
            run="poetry version patch && git add pyproject.toml",
            step=step,
        )

    @classmethod
    def step_commit_added_changes(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the commit changes step."""
        return cls.get_step(
            step_func=cls.step_commit_added_changes,
            run="git commit --no-verify -m '[skip ci] CI/CD: Committing possible added changes (e.g.: pyproject.toml and poetry.lock)'",  # noqa: E501
            step=step,
        )

    @classmethod
    def step_push_commits(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the push changes step."""
        return cls.get_step(
            step_func=cls.step_push_commits,
            run="git push",
            step=step,
        )

    @classmethod
    def step_create_and_push_tag(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the tag and push step."""
        return cls.get_step(
            step_func=cls.step_create_and_push_tag,
            run=f"git tag {cls.insert_version()} && git push origin {cls.insert_version()}",  # noqa: E501
            step=step,
        )

    @classmethod
    def step_create_folder(
        cls,
        *,
        folder: str,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the create folder step."""
        # should work on all OSs
        return cls.get_step(
            step_func=cls.step_create_folder,
            run=f"mkdir {folder}",
            step=step,
        )

    @classmethod
    def step_create_artifacts_folder(
        cls,
        *,
        folder: str = Builder.ARTIFACTS_DIR_NAME,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the create artifacts folder step."""
        return cls.step_create_folder(folder=folder, step=step)

    @classmethod
    def step_upload_artifacts(
        cls,
        *,
        name: str | None = None,
        path: str | Path = ARTIFACTS_DIR_NAME,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the upload artifacts step."""
        if name is None:
            name = cls.insert_artifact_name()
        return cls.get_step(
            step_func=cls.step_upload_artifacts,
            uses="actions/upload-artifact@main",
            with_={"name": name, "path": str(path)},
            step=step,
        )

    @classmethod
    def step_build_artifacts(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the build artifacts step."""
        return cls.get_step(
            step_func=cls.step_build_artifacts,
            run=get_poetry_run_pyrig_cli_cmd_script(build),
            step=step,
        )

    @classmethod
    def step_download_artifacts(
        cls,
        *,
        name: str | None = None,
        path: str | Path = ARTIFACTS_DIR_NAME,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the download artifacts step."""
        # omit name downloads all by default
        with_: dict[str, Any] = {"path": str(path)}
        if name is not None:
            with_["name"] = name
        with_["merge-multiple"] = "true"
        return cls.get_step(
            step_func=cls.step_download_artifacts,
            uses="actions/download-artifact@main",
            with_=with_,
            step=step,
        )

    @classmethod
    def step_build_changelog(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the build changelog step."""
        return cls.get_step(
            step_func=cls.step_build_changelog,
            uses="mikepenz/release-changelog-builder-action@develop",
            with_={"token": cls.insert_github_token()},
            step=step,
        )

    @classmethod
    def step_extract_version(
        cls,
        *,
        step: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Get the extract version step."""
        return cls.get_step(
            step_func=cls.step_extract_version,
            run=f'echo "version={cls.insert_version()}" >> $GITHUB_OUTPUT',
            step=step,
        )

    @classmethod
    def step_create_release(
        cls,
        *,
        step: dict[str, Any] | None = None,
        artifacts_pattern: str = ARTIFACTS_PATTERN,
    ) -> dict[str, Any]:
        """Get the create release step."""
        version = cls.insert_version_from_extract_version_step()
        return cls.get_step(
            step_func=cls.step_create_release,
            uses="ncipollo/release-action@main",
            with_={
                "tag": version,
                "name": f"{cls.insert_repository_name()} {version}",
                "body": cls.insert_changelog(),
                "artifacts": artifacts_pattern,
            },
            step=step,
        )

    # Insertions
    # ----------------------------------------------------------------------------
    @classmethod
    def insert_repo_token(cls) -> str:
        """Insert the repository token."""
        return "${{ secrets.REPO_TOKEN }}"

    @classmethod
    def insert_version(cls) -> str:
        """Insert the version."""
        return "v$(poetry version -s)"

    @classmethod
    def insert_version_from_extract_version_step(cls) -> str:
        """Insert the version from the extract version step."""
        # make dynamic with cls.make_id_from_func(cls.step_extract_version)
        return (
            "${{ "
            f"steps.{cls.make_id_from_func(cls.step_extract_version)}.outputs.version"
            " }}"
        )

    @classmethod
    def insert_changelog(cls) -> str:
        """Insert the changelog."""
        return (
            "${{ "
            f"steps.{cls.make_id_from_func(cls.step_build_changelog)}.outputs.changelog"
            " }}"
        )

    @classmethod
    def insert_github_token(cls) -> str:
        """Insert the GitHub token."""
        return "${{ secrets.GITHUB_TOKEN }}"

    @classmethod
    def insert_repository_name(cls) -> str:
        """Insert the repository name."""
        return "${{ github.event.repository.name }}"

    @classmethod
    def insert_ref_name(cls) -> str:
        """Insert the ref name."""
        return "${{ github.ref_name }}"

    @classmethod
    def insert_repository_ownwer(cls) -> str:
        """Insert the repository owner."""
        return "${{ github.repository_owner }}"

    @classmethod
    def insert_os(cls) -> str:
        """Insert the os."""
        return "${{ runner.os }}"

    @classmethod
    def insert_matrix_os(cls) -> str:
        """Insert the matrix os."""
        return "${{ matrix.os }}"

    @classmethod
    def insert_matrix_python_version(cls) -> str:
        """Insert the matrix python version."""
        return "${{ matrix.python-version }}"

    @classmethod
    def insert_artifact_name(cls) -> str:
        """Insert the artifact name."""
        return f"{get_src_package().__name__}-{cls.insert_os()}"

    # ifs
    @classmethod
    def if_matrix_is_os(cls, os: str) -> str:
        """Insert the matrix os."""
        return f"matrix.os == '{os}'"

    @classmethod
    def if_matrix_is_python_version(cls, python_version: str) -> str:
        """Insert the matrix python version."""
        return f"matrix.python-version == '{python_version}'"

    @classmethod
    def if_matrix_is_os_and_python_version(cls, os: str, python_version: str) -> str:
        """Insert the matrix os and python version."""
        return f"{cls.if_matrix_is_os(os)} && {cls.if_matrix_is_python_version(python_version)}"  # noqa: E501

    @classmethod
    def if_matrix_is_latest_python_version(cls) -> str:
        """Insert the matrix latest python version."""
        return cls.if_matrix_is_python_version(
            str(PyprojectConfigFile.get_latest_possible_python_version())
        )

    @classmethod
    def if_matrix_is_os_and_latest_python_version(cls, os: str) -> str:
        """Insert the matrix os and latest python version."""
        return cls.if_matrix_is_os_and_python_version(
            os, str(PyprojectConfigFile.get_latest_possible_python_version())
        )

    @classmethod
    def if_workflow_run_is_success(cls) -> str:
        """Insert the if workflow run is success."""
        return "${{ github.event.workflow_run.conclusion == 'success' }}"



"""Session-level test fixtures and utilities.

This module provides fixtures and test functions that operate at the session scope,
ensuring that project-wide conventions are followed and that the overall project
structure is correct. These fixtures are automatically applied to the test session
through pytest's autouse mechanism.
"""

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import pyrig
from pyrig.dev.configs.base.base import ConfigFile
from pyrig.dev.configs.pyproject import (
    PyprojectConfigFile,
)
from pyrig.src.modules.module import (
    import_module_with_default,
)
from pyrig.src.modules.package import (
    find_packages,
    get_src_package,
    walk_package,
)
from pyrig.src.os.os import run_subprocess
from pyrig.src.testing.assertions import assert_with_msg
from pyrig.src.testing.convention import (
    TESTS_PACKAGE_NAME,
    make_test_obj_importpath_from_obj,
    make_untested_summary_error_msg,
)
from pyrig.src.testing.create_tests import create_tests
from pyrig.src.testing.fixtures import autouse_session_fixture

if TYPE_CHECKING:
    from types import ModuleType

logger = logging.getLogger(__name__)


@autouse_session_fixture
def assert_dev_dependencies_config_is_correct() -> None:
    """Verify that the dev dependencies in consts.py are correct.

    This fixture runs once per test session and checks that the dev dependencies
    in consts.py are correct by comparing them to the dev dependencies in
    pyproject.toml.

    Raises:
        AssertionError: If the dev dependencies in consts.py are not correct

    """
    if PyprojectConfigFile.get_package_name() != pyrig.__name__:
        # this const is only used in pyrig
        # to be able to install them with setup.py
        return
    actual_dev_dependencies = PyprojectConfigFile.get_dev_dependencies()
    expected_dev_dependencies = PyprojectConfigFile.get_configs()["tool"]["poetry"][
        "group"
    ]["dev"]["dependencies"].keys()
    assert_with_msg(
        set(actual_dev_dependencies) == set(expected_dev_dependencies),
        "Dev dependencies in consts.py are not correct",
    )


@autouse_session_fixture
def assert_config_files_are_correct() -> None:
    """Verify that the dev dependencies are installed.

    This fixture runs once per test session and checks that the dev dependencies
    are installed by trying to import them.

    Raises:
        ImportError: If a dev dependency is not installed

    """
    subclasses = ConfigFile.get_all_subclasses()
    all_correct = all(subclass.is_correct() for subclass in subclasses)
    # subclasses of ConfigFile
    ConfigFile.init_config_files()

    assert_with_msg(
        all_correct,
        "Config files are not correct. Corrected the files. Please verify the changes.",
    )


@autouse_session_fixture
def assert_no_namespace_packages() -> None:
    """Verify that there are no namespace packages in the project.

    This fixture runs once per test session and checks that all packages in the
    project are regular packages with __init__.py files, not namespace packages.

    Raises:
        AssertionError: If any namespace packages are found

    """
    packages = find_packages(depth=None)
    namespace_packages = find_packages(depth=None, include_namespace_packages=True)

    any_namespace_packages = set(namespace_packages) - set(packages)
    assert_with_msg(
        not any_namespace_packages,
        f"Found namespace packages: {any_namespace_packages}. "
        f"All packages should have __init__.py files.",
    )


@autouse_session_fixture
def assert_all_src_code_in_one_package() -> None:
    """Verify that all source code is in a single package.

    This fixture runs once per test session and checks that there is only one
    source package besides the tests package.

    Raises:
        AssertionError: If there are multiple source packages

    """
    packages = find_packages(depth=0)
    src_package = get_src_package().__name__
    expected_packages = {TESTS_PACKAGE_NAME, src_package}
    assert_with_msg(
        set(packages) == expected_packages,
        f"Expected only packages {expected_packages}, but found {packages}",
    )


@autouse_session_fixture
def assert_src_package_correctly_named() -> None:
    """Verify that the source package is correctly named.

    This fixture runs once per test session and checks that the source package
    is correctly named after the project.

    Raises:
        AssertionError: If the source package is not correctly named

    """
    src_package = get_src_package().__name__
    config = PyprojectConfigFile
    expected_package = config.get_package_name()
    assert_with_msg(
        src_package == expected_package,
        f"Expected source package to be named {expected_package}, "
        f"but it is named {src_package}",
    )


@autouse_session_fixture
def assert_all_modules_tested() -> None:
    """Verify that the project structure is mirrored in tests.

    This fixture runs once per test session and checks that for every package and
    module in the source package, there is a corresponding test package and module.

    Raises:
        AssertionError: If any package or module doesn't have a corresponding test

    """
    src_package = get_src_package()

    # we will now go through all the modules in the src package and check
    # that there is a corresponding test module
    missing_tests_to_module: dict[str, ModuleType] = {}
    for package, modules in walk_package(src_package):
        test_package_name = make_test_obj_importpath_from_obj(package)
        test_package = import_module_with_default(test_package_name)
        if test_package is None:
            missing_tests_to_module[test_package_name] = package

        for module in modules:
            test_module_name = make_test_obj_importpath_from_obj(module)
            test_module = import_module_with_default(test_module_name)
            if test_module is None:
                missing_tests_to_module[test_module_name] = module

    if missing_tests_to_module:
        create_tests()

    msg = f"""Found missing tests. Tests skeletons were automatically created for:
    {make_untested_summary_error_msg(missing_tests_to_module.keys())}
"""
    assert_with_msg(
        not missing_tests_to_module,
        msg,
    )


@autouse_session_fixture
def assert_no_unit_test_package_usage() -> None:
    """Verify that the unit test package is not used in the project.

    This fixture runs once per test session and checks that the unit test package
    is not used in the project.

    Raises:
        AssertionError: If the unit test package is used

    """
    for path in Path().rglob("*.py"):
        assert_with_msg(
            "UnitTest".lower() not in path.read_text(encoding="utf-8"),
            f"Found unit test package usage in {path}. Use pytest instead.",
        )


@autouse_session_fixture
def assert_dependencies_are_up_to_date() -> None:
    """Verify that the dependencies are up to date.

    This fixture runs once per test session and runs poetry update --with dev
    to make sure the dependencies are up to date.
    """
    # update poetry itself
    completed_process = run_subprocess(["poetry", "self", "update"], check=True)
    stdout = completed_process.stdout.decode("utf-8")
    logger.info("Poetry self update output: %s", stdout)

    # update the dependencies
    completed_process = PyprojectConfigFile.update_dependencies(check=True)
    # if there were updates raise an error
    no_deps_updated_msg = "No dependencies to install or update"
    stdout = completed_process.stdout.decode("utf-8")

    msg = f"""This fixture runs poetry update --with dev to make sure the dependencies
    are up to date. The following dependencies were updated:
    {stdout}
"""
    assert_with_msg(
        no_deps_updated_msg in stdout,
        msg,
    )



"""Module-level test fixtures and utilities.

This module provides fixtures and test functions that operate at the module scope,
ensuring that all functions and classes within a module have corresponding tests.
These fixtures are automatically applied to all test modules through pytest's autouse
mechanism.
"""

from typing import TYPE_CHECKING

import pytest

from pyrig.src.testing.fixtures import autouse_module_fixture
from pyrig.src.testing.utils import assert_no_untested_objs

if TYPE_CHECKING:
    from types import ModuleType


@autouse_module_fixture
def assert_all_funcs_and_classes_tested(request: pytest.FixtureRequest) -> None:
    """Verify that all functions and classes in a module have corresponding tests.

    This fixture runs automatically for each test module and checks that every
    function and class defined in the corresponding source module has a test
    function or class defined in the test module.

    Args:
        request: The pytest fixture request object containing the current module

    Raises:
        AssertionError: If any function or class in the source module lacks a test

    """
    module: ModuleType = request.module
    assert_no_untested_objs(module)



"""Contains utilities for working with GitHub repositories."""

import logging
from typing import Any, Literal

from github import Github
from github.Auth import Token
from github.Repository import Repository

logger = logging.getLogger(__name__)

DEFAULT_BRANCH = "main"

DEFAULT_RULESET_NAME = f"{DEFAULT_BRANCH} protection"


def get_rules_payload(  # noqa: PLR0913
    *,
    creation: dict[str, Any] | None = None,
    update: dict[str, Any] | None = None,
    deletion: dict[str, Any] | None = None,
    required_linear_history: dict[str, Any] | None = None,
    merge_queue: dict[str, Any] | None = None,
    required_deployments: dict[str, Any] | None = None,
    required_signatures: dict[str, Any] | None = None,
    pull_request: dict[str, Any] | None = None,
    required_status_checks: dict[str, Any] | None = None,
    non_fast_forward: dict[str, Any] | None = None,
    commit_message_pattern: dict[str, Any] | None = None,
    commit_author_email_pattern: dict[str, Any] | None = None,
    committer_email_pattern: dict[str, Any] | None = None,
    branch_name_pattern: dict[str, Any] | None = None,
    tag_name_pattern: dict[str, Any] | None = None,
    file_path_restriction: dict[str, Any] | None = None,
    max_file_path_length: dict[str, Any] | None = None,
    file_extension_restriction: dict[str, Any] | None = None,
    max_file_size: dict[str, Any] | None = None,
    workflows: dict[str, Any] | None = None,
    code_scanning: dict[str, Any] | None = None,
    copilot_code_review: dict[str, Any] | None = None,
) -> list[dict[str, Any]]:
    """Build a rules array for a GitHub ruleset.

    Args:
        creation: Only allow users with bypass permission to create matching
            refs.
        update: Only allow users with bypass permission to update matching
            refs.
        deletion: Only allow users with bypass permissions to delete matching
            refs.
        required_linear_history: Prevent merge commits from being pushed to
            matching refs.
        merge_queue: Merges must be performed via a merge queue.
        required_deployments: Choose which environments must be successfully
            deployed to before refs can be pushed.
        required_signatures: Commits pushed to matching refs must have verified
            signatures.
        pull_request: Require all commits be made to a non-target branch and
            submitted via a pull request.
        required_status_checks: Choose which status checks must pass before the
            ref is updated.
        non_fast_forward: Prevent users with push access from force pushing to
            refs.
        commit_message_pattern: Parameters to be used for the
            commit_message_pattern rule.
        commit_author_email_pattern: Parameters to be used for the
            commit_author_email_pattern rule.
        committer_email_pattern: Parameters to be used for the
            committer_email_pattern rule.
        branch_name_pattern: Parameters to be used for the branch_name_pattern
            rule.
        tag_name_pattern: Parameters to be used for the tag_name_pattern rule.
        file_path_restriction: Prevent commits that include changes in
            specified file and folder paths.
        max_file_path_length: Prevent commits that include file paths that
            exceed the specified character limit.
        file_extension_restriction: Prevent commits that include files with
            specified file extensions.
        max_file_size: Prevent commits with individual files that exceed the
            specified limit.
        workflows: Require all changes made to a targeted branch to pass the
            specified workflows.
        code_scanning: Choose which tools must provide code scanning results
            before the reference is updated.
        copilot_code_review: Request Copilot code review for new pull requests
            automatically.

    Returns:
        A list of rule objects to be used in a GitHub ruleset.
    """
    rules: list[dict[str, Any]] = []

    rule_map = {
        "creation": creation,
        "update": update,
        "deletion": deletion,
        "required_linear_history": required_linear_history,
        "merge_queue": merge_queue,
        "required_deployments": required_deployments,
        "required_signatures": required_signatures,
        "pull_request": pull_request,
        "required_status_checks": required_status_checks,
        "non_fast_forward": non_fast_forward,
        "commit_message_pattern": commit_message_pattern,
        "commit_author_email_pattern": commit_author_email_pattern,
        "committer_email_pattern": committer_email_pattern,
        "branch_name_pattern": branch_name_pattern,
        "tag_name_pattern": tag_name_pattern,
        "file_path_restriction": file_path_restriction,
        "max_file_path_length": max_file_path_length,
        "file_extension_restriction": file_extension_restriction,
        "max_file_size": max_file_size,
        "workflows": workflows,
        "code_scanning": code_scanning,
        "copilot_code_review": copilot_code_review,
    }

    for rule_type, rule_config in rule_map.items():
        if rule_config is not None:
            rule_obj: dict[str, Any] = {"type": rule_type}
            if rule_config:  # If there are parameters
                rule_obj["parameters"] = rule_config
            rules.append(rule_obj)

    return rules


def create_or_update_ruleset(  # noqa: PLR0913
    token: str,
    owner: str,
    repo_name: str,
    *,
    ruleset_name: str,
    enforcement: Literal["active", "disabled", "evaluate"] = "active",
    target: Literal["branch", "tag", "push"] = "branch",
    bypass_actors: list[dict[str, Any]] | None = None,
    conditions: dict[
        Literal["ref_name"], dict[Literal["include", "exclude"], list[str]]
    ]
    | None = None,
    rules: list[dict[str, Any]] | None = None,
) -> Any:
    """Create a ruleset for the repository."""
    repo = get_repo(token, owner, repo_name)
    ruleset_id = ruleset_exists(
        token=token, owner=owner, repo_name=repo_name, ruleset_name=ruleset_name
    )
    method = "PUT" if ruleset_id else "POST"
    url = f"{repo.url}/rulesets"

    if ruleset_id:
        url += f"/{ruleset_id}"

    payload: dict[str, Any] = {
        "name": ruleset_name,
        "enforcement": enforcement,
        "target": target,
        "conditions": conditions,
        "rules": rules,
    }
    if bypass_actors:
        payload["bypass_actors"] = bypass_actors

    _headers, res = repo._requester.requestJsonAndCheck(  # noqa: SLF001
        method,
        url,
        headers={
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": "2022-11-28",
        },
        input=payload,
    )

    return res


def get_all_rulesets(token: str, owner: str, repo_name: str) -> Any:
    """Get all rulesets for the repository."""
    repo = get_repo(token, owner, repo_name)
    url = f"{repo.url}/rulesets"
    method = "GET"
    _headers, res = repo._requester.requestJsonAndCheck(  # noqa: SLF001
        method,
        url,
        headers={
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": "2022-11-28",
        },
    )
    return res


def get_repo(token: str, owner: str, repo_name: str) -> Repository:
    """Get the repository."""
    auth = Token(token)
    github = Github(auth=auth)
    return github.get_repo(f"{owner}/{repo_name}")


def ruleset_exists(token: str, owner: str, repo_name: str, ruleset_name: str) -> int:
    """Check if the main protection ruleset exists."""
    rulesets = get_all_rulesets(token, owner, repo_name)
    main_ruleset = next((rs for rs in rulesets if rs["name"] == ruleset_name), None)
    return main_ruleset["id"] if main_ruleset else 0



"""Config utilities for py.typed."""

from pathlib import Path

from pyrig.dev.configs.base.base import TypedConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile


class PyTypedConfigFile(TypedConfigFile):
    """Config file for py.typed."""

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path(PyprojectConfigFile.get_package_name())



"""A script that can be called after you installed the package.

This script calls create tests, creates the pre-commit config, and
creates the pyproject.toml file and some other things to set up a project.
This package assumes you are using poetry and pre-commit.
This script is intended to be called once at the beginning of a project.
"""

import logging
from collections.abc import Callable
from typing import Any

from pyrig.dev.configs.git.pre_commit import PreCommitConfigConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile
from pyrig.dev.configs.testing.conftest import ConftestConfigFile
from pyrig.src.project.create_root import create_root
from pyrig.src.testing.create_tests import create_tests

logger = logging.getLogger(__name__)


SETUP_STEPS: list[Callable[..., Any]] = [
    create_root,
    create_tests,
    PyprojectConfigFile.update_dependencies,
    PreCommitConfigConfigFile.run_hooks,
    ConftestConfigFile.run_tests,
    PyprojectConfigFile.install_dependencies,  # to activate cli
]


def init() -> None:
    """Set up the project."""
    for step in SETUP_STEPS:
        logger.info("Running setup step: %s", step.__name__)
        step()
    logger.info("Setup complete!")



"""Contains a dict with the dev dependencies.

For poetry when pyrig is a dependency.
pyrigwill add these automatically to the pyproject.toml file.
pyrig's PyprojectConfigFile will auto dump the config here so it can access it
when being a dependency in another project.
"""

DEV_DEPENDENCIES: dict[str, str | dict[str, str]] = {
    "ruff": "*",
    "pre-commit": "*",
    "mypy": "*",
    "pytest": "*",
    "bandit": "*",
    "types-setuptools": "*",
    "types-tqdm": "*",
    "types-defusedxml": "*",
    "types-pyyaml": "*",
    "pytest-mock": "*",
    "types-networkx": "*",
    "types-pyinstaller": "*",
    "pyinstaller": {"version": "*", "python": "<3.15"},
}



"""Configs for pyrig.

All subclasses of ConfigFile in the configs package are automatically called.
"""

from types import ModuleType

from pyrig.dev.configs import configs
from pyrig.dev.configs.base.base import CopyModuleConfigFile


class ConfigsConfigFile(CopyModuleConfigFile):
    """Config file for configs.py."""

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return configs



"""Package-level test fixtures and utilities.

This module provides fixtures and test functions that operate at the package scope,
ensuring that all modules within a package have corresponding tests.
These fixtures are automatically applied to all test packages through pytest's autouse
mechanism.
"""



"""Function utilities for introspection and manipulation.

This module provides utility functions for working with Python functions,
including extracting functions from modules and manipulating function objects.
These utilities are particularly useful for reflection, testing, and
dynamic code generation.
"""

import inspect
from collections.abc import Callable
from importlib import import_module
from types import ModuleType
from typing import Any

from pyrig.src.modules.inspection import get_def_line, get_obj_members


def is_func_or_method(obj: Any) -> bool:
    """Return True if *obj* is a function or method.

    This function checks if the given object is a function or method,
    including those defined in a class body.

    Args:
        obj: The object to check

    Returns:
        bool: True if the object is a function or method, False otherwise

    """
    return inspect.isfunction(obj) or inspect.ismethod(obj)


def is_func(obj: Any) -> bool:
    """Return True if *obj* is a 'method-like' attribute as it appears in a class body.

    Accepts:


        • plain functions (instance methods)
        • staticmethod / classmethod descriptors
        • property descriptors (getter counts as method)
        • decorated functions that keep a __wrapped__ chain

    Returns:
        bool: True if the object is a method-like attribute, False otherwise

    """
    # plain function

    if is_func_or_method(obj):
        return True

    if isinstance(obj, (staticmethod, classmethod, property)):
        return True

    unwrapped = inspect.unwrap(obj)

    return is_func_or_method(unwrapped)


def get_all_functions_from_module(
    module: ModuleType | str, *, include_annotate: bool = False
) -> list[Callable[..., Any]]:
    """Get all functions defined in a module.

    Retrieves all function objects that are defined directly in the specified module,
    excluding imported functions.
    The functions are sorted by their line number in the module.

    Args:
        module: The module to extract functions from
        include_annotate: If False, exclude __annotate__ method
        introduced in Python 3.14, defaults to False

    Returns:
        A list of callable functions defined in the module

    """
    from pyrig.src.modules.module import (  # noqa: PLC0415  # avoid circular import
        get_module_of_obj,
    )

    if isinstance(module, str):
        module = import_module(module)
    funcs = [
        func
        for _name, func in get_obj_members(module, include_annotate=include_annotate)
        if is_func(func)
        if get_module_of_obj(func).__name__ == module.__name__
    ]
    # sort by definition order
    return sorted(funcs, key=get_def_line)


def unwrap_method(method: Any) -> Callable[..., Any] | Any:
    """Unwrap a method to its underlying function.

    Args:
        method: The method to unwrap

    Returns:
        The underlying function of the method

    """
    if isinstance(method, (staticmethod, classmethod)):
        method = method.__func__
    if isinstance(method, property):
        method = method.fget
    return inspect.unwrap(method)


def is_abstractmethod(method: Any) -> bool:
    """Check if a method is an abstract method.

    Args:
        method: The method to check

    Returns:
        True if the method is an abstract method, False otherwise

    """
    method = unwrap_method(method)
    return getattr(method, "__isabstractmethod__", False)



"""Config utilities for LICENSE."""

from pathlib import Path

from pyrig.dev.configs.base.base import TextConfigFile


class LicenceConfigFile(TextConfigFile):
    """Config file for LICENSE."""

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return "LICENSE"

    @classmethod
    def get_path(cls) -> Path:
        """Get the path to the config file."""
        return Path(cls.get_filename())

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return ""

    @classmethod
    def get_content_str(cls) -> str:
        """Get the content."""
        return ""



"""Utilities for working with Python projects."""

from pyrig.dev.configs.base.base import ConfigFile


def create_root() -> None:
    """Create the project root."""
    ConfigFile.init_config_files()



"""Function-level test fixtures and utilities.

This module provides fixtures and test functions that operate at the function scope,
ensuring proper setup and teardown for individual test functions.
These fixtures are automatically applied to all test functions through pytest's autouse
mechanism.
"""



"""Has config utilities for pre-commit."""

import logging
from pathlib import Path
from typing import Any

from pyrig.dev.configs.base.base import YamlConfigFile
from pyrig.src.os.os import run_subprocess
from pyrig.src.project.poetry.poetry import (
    POETRY_ARG,
    POETRY_RUN_ARGS,
    get_script_from_args,
)

logger = logging.getLogger(__name__)


class PreCommitConfigConfigFile(YamlConfigFile):
    """Config file for pre-commit."""

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        filename = super().get_filename()
        return f".{filename.replace('_', '-')}"

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_hook(
        cls,
        name: str,
        args: list[str],
        *,
        language: str = "system",
        always_run: bool = True,
        pass_filenames: bool = False,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Get a hook."""
        if args[0] != POETRY_ARG:
            args = POETRY_RUN_ARGS + args
        return {
            "id": name,
            "name": name,
            "entry": get_script_from_args(args),
            "language": language,
            "always_run": always_run,
            "pass_filenames": pass_filenames,
            **kwargs,
        }

    @classmethod
    def get_configs(cls) -> dict[str, Any]:
        """Get the config."""
        hooks: list[dict[str, Any]] = [
            cls.get_hook(
                "check-package-manager-config",
                ["poetry", "check", "--strict"],
            ),
            cls.get_hook(
                "install-dependencies",
                ["poetry", "install", "--with", "dev"],
            ),
            cls.get_hook(
                "create-root",
                ["pyrig", "create-root"],
            ),
            cls.get_hook(
                "lint-code",
                ["ruff", "check", "--fix"],
            ),
            cls.get_hook(
                "format-code",
                ["ruff", "format"],
            ),
            cls.get_hook(
                "check-static-types",
                ["mypy", "--exclude-gitignore"],
            ),
            cls.get_hook(
                "check-security",
                ["bandit", "-c", "pyproject.toml", "-r", "."],
            ),
        ]
        return {
            "repos": [
                {
                    "repo": "local",
                    "hooks": hooks,
                },
            ]
        }

    def __init__(self) -> None:
        """Init the file."""
        super().__init__()

    @classmethod
    def install(cls) -> None:
        """Installs the pre commits in the config."""
        logger.info("Running pre-commit install")
        run_subprocess(["pre-commit", "install"])

    @classmethod
    def run_hooks(
        cls,
        *,
        with_install: bool = True,
        all_files: bool = True,
        add_before_commit: bool = True,
        verbose: bool = True,
        check: bool = True,
    ) -> None:
        """Runs the pre-commit hooks."""
        if add_before_commit:
            logger.info("Adding all files to git")
            run_subprocess(["git", "add", "."])
        if with_install:
            cls.install()
        logger.info("Running pre-commit run")
        args = ["pre-commit", "run"]
        if all_files:
            args.append("--all-files")
        if verbose:
            args.append("--verbose")
        run_subprocess([*args], check=check)



"""Pytest configuration for pirig tests.

finds all the plugins in the tests directory and the package's testing module
and adds them to pytest_plugins. This way defining reusable fixtures is easy.
"""

from pathlib import Path

import pyrig
from pyrig import dev
from pyrig.src.modules.module import to_module_name, to_path

tests_to_fixtures = "tests.base.fixtures"
tests_to_fixtures_path = to_path(tests_to_fixtures, is_package=True)

package_root_path = Path(pyrig.__file__).parent.parent
package_plugin_path = Path(dev.__file__).parent / tests_to_fixtures_path
package_plugin_module_names = [
    to_module_name(path.relative_to(package_root_path))
    for path in package_plugin_path.rglob("*.py")
]

if not package_plugin_module_names or not package_plugin_path.exists():
    msg = f"Found no plugins in {package_plugin_path}"
    raise ValueError(msg)


custom_plugin_path = tests_to_fixtures_path
custom_plugin_module_names = [
    to_module_name(path) for path in custom_plugin_path.rglob("*.py")
]
if not custom_plugin_module_names and custom_plugin_path.exists():
    msg = f"Found no plugins in {custom_plugin_path}"
    raise ValueError(msg)

pytest_plugins = [
    *package_plugin_module_names,
    *custom_plugin_module_names,
]



"""Contains the release workflow.

This workflow is used to create a release on GitHub.
"""

from typing import Any

from pyrig.dev.artifacts.builder.base.base import Builder
from pyrig.dev.configs.workflows.health_check import HealthCheckWorkflow


class ReleaseWorkflow(HealthCheckWorkflow):
    """Release workflow.

    This workflow is triggered by a push to the main branch.
    It creates a tag for the release and builds a changelog.
    With tag and changelog it creates a release on GitHub
    """

    @classmethod
    def get_workflow_triggers(cls) -> dict[str, Any]:
        """Get the workflow triggers."""
        triggers = super().get_workflow_triggers()
        del triggers["pull_request"]
        triggers.update(cls.on_push())
        triggers.update(cls.on_schedule(cron="0 6 * * 2"))
        return triggers

    @classmethod
    def get_permissions(cls) -> dict[str, Any]:
        """Get the workflow permissions."""
        permissions = super().get_permissions()
        permissions["contents"] = "write"
        return permissions

    @classmethod
    def get_jobs(cls) -> dict[str, Any]:
        """Get the workflow jobs."""
        jobs = super().get_jobs()
        last_job_name = list(jobs.keys())[-1]
        jobs.update(cls.job_build(needs=[last_job_name]))
        jobs.update(cls.job_release())
        return jobs

    @classmethod
    def job_build(cls, needs: list[str] | None = None) -> dict[str, Any]:
        """Get the build job."""
        return cls.get_job(
            job_func=cls.job_build,
            needs=needs,
            strategy=cls.strategy_matrix_os(),
            runs_on=cls.insert_matrix_os(),
            steps=cls.steps_build(),
        )

    @classmethod
    def job_release(cls) -> dict[str, Any]:
        """Get the release job."""
        return cls.get_job(
            job_func=cls.job_release,
            needs=[cls.make_id_from_func(cls.job_build)],
            steps=cls.steps_release(),
        )

    @classmethod
    def steps_build(cls) -> list[dict[str, Any]]:
        """Get the build steps."""
        non_abstract_builders = Builder.get_non_abstract_subclasses()
        if not non_abstract_builders:
            return [cls.step_no_builder_defined()]
        return [
            *cls.steps_core_matrix_setup(),
            cls.step_build_artifacts(),
            cls.step_upload_artifacts(),
        ]

    @classmethod
    def steps_release(cls) -> list[dict[str, Any]]:
        """Get the release steps."""
        return [
            *cls.steps_core_installed_setup(repo_token=True),
            cls.step_setup_git(),
            cls.step_add_version_patch(),
            cls.step_run_pre_commit_hooks(),
            cls.step_commit_added_changes(),
            cls.step_push_commits(),
            cls.step_create_and_push_tag(),
            cls.step_extract_version(),
            cls.step_download_artifacts(),
            cls.step_build_changelog(),
            cls.step_create_release(),
        ]



"""GitHub utilities for working with GitHub repositories."""

import os

from pyrig.dev.configs.dot_env import DotEnvConfigFile


def get_github_repo_token() -> str:
    """Get the GitHub token."""
    # try os env first
    token = os.getenv("REPO_TOKEN")
    if token:
        return token

    # try .env next
    dotenv_path = DotEnvConfigFile.get_path()
    if not dotenv_path.exists():
        msg = f"Expected {dotenv_path} to exist"
        raise ValueError(msg)
    dotenv = DotEnvConfigFile.load()
    token = dotenv.get("REPO_TOKEN")
    if token:
        return token

    msg = f"Expected REPO_TOKEN in {dotenv_path}"
    raise ValueError(msg)


def running_in_github_actions() -> bool:
    """Check if we are running in a GitHub action."""
    return os.getenv("GITHUB_ACTIONS", "false") == "true"



"""Config utilities for subcommands.py."""

from types import ModuleType

from pyrig.dev.cli import subcommands
from pyrig.dev.configs.base.base import CopyModuleConfigFile


class SubcommandsConfigFile(CopyModuleConfigFile):
    """Config file for subcommands.py."""

    @classmethod
    def get_src_module(cls) -> ModuleType:
        """Get the source module."""
        return subcommands

    @classmethod
    def get_content_str(cls) -> str:
        """Get the content."""
        content = super().get_content_str()
        # override content bc we have a subcommnds
        parts = content.split('"""', 2)
        return '"""' + parts[1] + '"""\n'



"""Config File for README.md."""

from pathlib import Path

import pyrig
from pyrig.dev.configs.base.base import TextConfigFile
from pyrig.dev.configs.pyproject import PyprojectConfigFile


class ReadmeConfigFile(TextConfigFile):
    """Config file for README.md."""

    @classmethod
    def get_filename(cls) -> str:
        """Get the filename of the config file."""
        return "README"

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_file_extension(cls) -> str:
        """Get the file extension of the config file."""
        return "md"

    @classmethod
    def get_content_str(cls) -> str:
        """Get the content."""
        content = f"""# {PyprojectConfigFile.get_project_name()}
"""
        if PyprojectConfigFile.get_project_name() != pyrig.__name__:
            content += f"""
(This project uses [{PyprojectConfigFile.get_project_name_from_pkg_name(pyrig.__name__)}](https://github.com/Winipedia/{pyrig.__name__}))
"""
        return content



"""Contains the publish workflow.

This workflow is used to publish the package to PyPI with poetry.
"""

from typing import Any

from pyrig.dev.configs.workflows.base.base import Workflow
from pyrig.dev.configs.workflows.release import ReleaseWorkflow


class PublishWorkflow(Workflow):
    """Publish workflow.

    This workflow is triggered by the release workflow.
    It publishes the package to PyPI with poetry.
    """

    @classmethod
    def get_workflow_triggers(cls) -> dict[str, Any]:
        """Get the workflow triggers."""
        triggers = super().get_workflow_triggers()
        triggers.update(
            cls.on_workflow_run(workflows=[ReleaseWorkflow.get_workflow_name()])
        )
        return triggers

    @classmethod
    def get_jobs(cls) -> dict[str, Any]:
        """Get the workflow jobs."""
        jobs: dict[str, Any] = {}
        jobs.update(cls.job_publish())
        return jobs

    @classmethod
    def job_publish(cls) -> dict[str, Any]:
        """Get the publish job."""
        return cls.get_job(
            job_func=cls.job_publish,
            steps=cls.steps_publish(),
            if_condition=cls.if_workflow_run_is_success(),
        )

    @classmethod
    def steps_publish(cls) -> list[dict[str, Any]]:
        """Get the publish steps."""
        return [
            *cls.steps_core_setup(),
            cls.step_add_pypi_token_to_poetry(),
            cls.step_publish_to_pypi(),
        ]



"""Config utilities for poetry and pyproject.toml."""

from functools import cache
from pathlib import Path
from subprocess import CompletedProcess  # nosec: B404
from typing import Any, cast

import requests
from packaging.version import Version

from pyrig.dev.configs.base.base import TomlConfigFile
from pyrig.dev.configs.python.experiment import ExperimentConfigFile
from pyrig.src.os.os import run_subprocess
from pyrig.src.project.poetry.dev_deps import DEV_DEPENDENCIES
from pyrig.src.project.poetry.versions import VersionConstraint
from pyrig.src.testing.convention import TEST_MODULE_PREFIX, TESTS_PACKAGE_NAME


class PyprojectConfigFile(TomlConfigFile):
    """Config file for pyproject.toml."""

    @classmethod
    def dump(cls, config: dict[str, Any] | list[Any]) -> None:
        """Dump the config file.

        We remove the wrong dependencies from the config before dumping.
        So we do not want dependencies under tool.poetry.dependencies but
        under project.dependencies. And we do not want dev dependencies under
        tool.poetry.dev-dependencies but under tool.poetry.group.dev.dependencies.
        """
        if not isinstance(config, dict):
            msg = f"Cannot dump {config} to pyproject.toml file."
            raise TypeError(msg)
        config = cls.remove_wrong_dependencies(config)
        super().dump(config)

    @classmethod
    def get_parent_path(cls) -> Path:
        """Get the path to the config file."""
        return Path()

    @classmethod
    def get_project_name_from_cwd(cls) -> str:
        """Get the repository name.

        Is the parent folder the project ives in and should be the same as the
        project name.
        """
        cwd = Path.cwd()
        return cwd.name

    @classmethod
    def get_pkg_name_from_cwd(cls) -> str:
        """Get the package name from the cwd."""
        return cls.get_pkg_name_from_project_name(cls.get_project_name_from_cwd())

    @classmethod
    def get_project_description(cls) -> str:
        """Get the project description."""
        return str(cls.load().get("project", {}).get("description", ""))

    @classmethod
    def get_configs(cls) -> dict[str, Any]:
        """Get the config."""
        from pyrig.dev.cli import (  # noqa: PLC0415
            cli,
        )

        return {
            "project": {
                "name": cls.get_project_name_from_cwd(),
                "readme": "README.md",
                "dynamic": ["dependencies"],
                "scripts": {
                    cls.get_project_name(): f"{cli.__name__}:{cli.main.__name__}"
                },
            },
            "build-system": {
                "requires": ["poetry-core>=2.0.0,<3.0.0"],
                "build-backend": "poetry.core.masonry.api",
            },
            "tool": {
                "poetry": {
                    "packages": [
                        {
                            "include": cls.get_pkg_name_from_cwd(),
                        }
                    ],
                    "dependencies": cls.make_dependency_to_version_dict(
                        cls.get_dependencies()
                    ),
                    "group": {
                        "dev": {
                            "dependencies": cls.make_dependency_to_version_dict(
                                cls.get_dev_dependencies(),
                                additional=DEV_DEPENDENCIES,
                            )
                        }
                    },
                },
                "ruff": {
                    "exclude": [".*", "**/migrations/*.py"],
                    "lint": {
                        "select": ["ALL"],
                        "ignore": ["D203", "D213", "COM812", "ANN401"],
                        "fixable": ["ALL"],
                        "per-file-ignores": {
                            f"{TESTS_PACKAGE_NAME}/**/*.py": ["S101"],
                        },
                        "pydocstyle": {"convention": "google"},
                    },
                },
                "mypy": {
                    "strict": True,
                    "warn_unreachable": True,
                    "show_error_codes": True,
                    "files": ".",
                },
                "pytest": {"ini_options": {"testpaths": [TESTS_PACKAGE_NAME]}},
                "bandit": {
                    "exclude_dirs": ["./" + ExperimentConfigFile.get_path().as_posix()],
                    "assert_used": {
                        "skips": [f"*{TEST_MODULE_PREFIX}*.py"],
                    },
                },
            },
        }

    @classmethod
    def make_dependency_to_version_dict(
        cls,
        dependencies: dict[str, str | dict[str, str]],
        additional: dict[str, str | dict[str, str]] | None = None,
    ) -> dict[str, str | dict[str, str]]:
        """Make a dependency to version dict.

        Args:
            dependencies: Dependencies to add
            additional: Additional dependencies to add

        Returns:
            Dependency to version dict
        """
        if additional is None:
            additional = {}
        dependencies.update(additional)
        dep_to_version_dict: dict[str, str | dict[str, str]] = {}
        for dep, version in dependencies.items():
            at_file_dep = " @ file://"
            if at_file_dep in dep:
                dep_new, path = dep.split(at_file_dep)
                dep_to_version_dict[dep_new] = {"path": path}
                continue
            if isinstance(version, dict):
                dep_to_version_dict[dep] = version
                continue
            dep_to_version_dict[dep] = "*"
        return dep_to_version_dict

    @classmethod
    def get_package_name(cls) -> str:
        """Get the package name."""
        project_name = cls.get_project_name()
        return cls.get_pkg_name_from_project_name(project_name)

    @classmethod
    def get_pkg_name_from_project_name(cls, project_name: str) -> str:
        """Get the package name from the project name."""
        return project_name.replace("-", "_")

    @classmethod
    def get_project_name_from_pkg_name(cls, pkg_name: str) -> str:
        """Get the project name from the package name."""
        return pkg_name.replace("_", "-")

    @classmethod
    def get_project_name(cls) -> str:
        """Get the project name."""
        return str(cls.load().get("project", {}).get("name", ""))

    @classmethod
    def remove_wrong_dependencies(cls, config: dict[str, Any]) -> dict[str, Any]:
        """Remove the wrong dependencies from the config."""
        # raise if the right sections do not exist
        if config.get("tool", {}).get("poetry", {}).get("dependencies") is None:
            msg = "No dependencies section in config"
            raise ValueError(msg)

        if (
            config.get("tool", {}).get("poetry", {}).get("group", {}).get("dev", {})
            is None
        ):
            msg = "No dev dependencies section in config"
            raise ValueError(msg)

        # remove the wrong dependencies sections if they exist
        if config.get("project", {}).get("dependencies") is not None:
            del config["project"]["dependencies"]
        if config.get("tool", {}).get("poetry", {}).get("dev-dependencies") is not None:
            del config["tool"]["poetry"]["dev-dependencies"]

        return config

    @classmethod
    def get_all_dependencies(cls) -> dict[str, str | dict[str, str]]:
        """Get all dependencies."""
        all_deps = cls.get_dependencies()
        all_deps.update(cls.get_dev_dependencies())
        return all_deps

    @classmethod
    def get_dev_dependencies(cls) -> dict[str, str | dict[str, str]]:
        """Get the dev dependencies."""
        dev_deps: dict[str, str | dict[str, str]] = (
            cls.load().get("tool", {}).get("poetry", {}).get("dev-dependencies", {})
        )
        tool_dev_deps = (
            cls.load()
            .get("tool", {})
            .get("poetry", {})
            .get("group", {})
            .get("dev", {})
            .get("dependencies", {})
        )
        dev_deps.update(tool_dev_deps)
        return dev_deps

    @classmethod
    def get_dependencies(cls) -> dict[str, str | dict[str, str]]:
        """Get the dependencies."""
        deps_raw = set(cls.load().get("project", {}).get("dependencies", {}))
        deps = {
            d.split("(")[0].strip(): d.split("(")[1].split(")")[0].strip()
            if "(" in d
            else "*"
            for d in deps_raw
        }

        tool_deps = cls.load().get("tool", {}).get("poetry", {}).get("dependencies", {})
        deps.update(tool_deps)
        return deps

    @classmethod
    def get_authors(cls) -> list[dict[str, str]]:
        """Get the authors."""
        return cast(
            "list[dict[str, str]]", cls.load().get("project", {}).get("authors", [])
        )

    @classmethod
    def get_main_author(cls) -> dict[str, str]:
        """Get the main author.

        Assumes the main author is the first author.
        """
        return cls.get_authors()[0]

    @classmethod
    def get_main_author_name(cls) -> str:
        """Get the main author name."""
        return cls.get_main_author()["name"]

    @classmethod
    @cache
    def fetch_latest_python_version(cls) -> Version:
        """Fetch the latest python version from python.org."""
        url = "https://endoflife.date/api/python.json"
        resp = requests.get(url, timeout=10)
        resp.raise_for_status()
        data = resp.json()
        # first element has metadata for latest stable
        latest_version = data[0]["latest"]
        return Version(latest_version)

    @classmethod
    def get_latest_possible_python_version(cls) -> Version:
        """Get the latest possible python version."""
        constraint = cls.load()["project"]["requires-python"]
        version_constraint = VersionConstraint(constraint)
        version = version_constraint.get_upper_inclusive()
        if version is None:
            version = cls.fetch_latest_python_version()
        return version

    @classmethod
    def get_first_supported_python_version(cls) -> Version:
        """Get the first supported python version."""
        constraint = cls.load()["project"]["requires-python"]
        version_constraint = VersionConstraint(constraint)
        lower = version_constraint.get_lower_inclusive()
        if lower is None:
            msg = "Need a lower bound for python version"
            raise ValueError(msg)
        return lower

    @classmethod
    def get_supported_python_versions(cls) -> list[Version]:
        """Get all supported python versions."""
        constraint = cls.load()["project"]["requires-python"]
        version_constraint = VersionConstraint(constraint)
        return version_constraint.get_version_range(
            level="minor", upper_default=cls.fetch_latest_python_version()
        )

    @classmethod
    def update_dependencies(cls, *, check: bool = True) -> CompletedProcess[bytes]:
        """Update the dependencies."""
        return run_subprocess(["poetry", "update", "--with", "dev"], check=check)

    @classmethod
    def install_dependencies(cls, *, check: bool = True) -> CompletedProcess[bytes]:
        """Install the dependencies."""
        return run_subprocess(["poetry", "install", "--with", "dev"], check=check)
